import numpy
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.image as image
import matplotlib.animation as animation
import matplotlib.cm as cm
import copy
import collections.abc
import math

def imv_normalize(v):
  return min(max(0.0, float(v)), 1.0)

def image_preprocess(data):
  for i in range(0,len(data)):
    for j in range(0,len(data[i])):
      if isinstance(data[i][j], collections.abc.Sequence):
        if len(data[i][j]) < 3:
            raise ValueError("color pixel must be an array of length 3")
        for k in range(0,3):
          data[i][j][k] = imv_normalize(data[i][j][k])
      else:
        if not (type(data[i][j]) is int or type(data[i][j]) is float):
          raise TypeError("monochrome image must be a 2d-array of numbers")
        v = imv_normalize(data[i][j])
        data[i][j] = [v, v, v]

def image_show(data):
    if not (isinstance(data, collections.abc.Sequence) and
            isinstance(data[0], collections.abc.Sequence)):
        raise TypeError("image_show: argument must be an image (i.e., a 2d-array)")
    tmp = copy.deepcopy(data)
    image_preprocess(tmp)
    picture = plt.imshow(tmp)
    plt.show()
    return picture

def animation_show(data, interval=500):
    if not (isinstance(data, collections.abc.Sequence) and
            isinstance(data[0], collections.abc.Sequence) and
            isinstance(data[0][0], collections.abc.Sequence)):
        raise TypeError("animation_show: argument must be an array of images (i.e., a 2d-array)")
    fig = plt.figure()
    a = []
    for i in range(0,len(data)):
        tmp = copy.deepcopy(data[i])
        image_preprocess(tmp)
        a.append([plt.imshow(tmp)])
    matplotlib.rc('animation', html='jshtml')
    ani = animation.ArtistAnimation(fig, a, interval=interval)
    try:
      import google.colab
      return ani
    except:
      plt.show()
      return ani

def linear_fit(data):
  if not isinstance(data, collections.abc.Sequence):
    TypeError("linear_fit: argument must be an array")
  xdata = [i[0] for i in data]
  ydata = [i[1] for i in data]
  coef, stats = numpy.polynomial.polynomial.polyfit(xdata, ydata, 1, full=True)
  fitp = numpy.polynomial.polynomial.Polynomial(coef)
  ma = math.ceil(max(xdata))
  mi = math.floor(min(xdata))
  xp = numpy.linspace(mi, ma, (ma - mi))
  plt.plot(xdata, ydata, '.', xp, fitp(xp), '-')
  plt.show()
  return stats[0][0]

def plotdata(data, line=False, multiplot=False, **kwargs):
  if line:
    pltfun = plt.plot
  else:
    pltfun = plt.scatter
  if not isinstance(data, collections.abc.Sequence):
    TypeError("plotdata: argument must be an array")
  if multiplot:
    if not isinstance(data[0], collections.abc.Sequence):
      TypeError("plotdata: argument must be a 2D-array if multiline=True")
    for d in data:
        pltfun(list(range(len(d))), d, **kwargs)
  else:
    if isinstance(data[0], collections.abc.Sequence):
      if isinstance(data[0][0], collections.abc.Sequence):
        plot_clusters(data)
      else:
        pltfun([i[0] for i in data], [i[1] for i in data], **kwargs)
    else:
      pltfun(list(range(len(data))), data, **kwargs)
  plt.show()

def plot_clusters(data, **kwargs):
  cnum = len(data)
  n = 0
  for x in data:
    n = n + len(x)
  xdata = [0] * n
  ydata = [0] * n
  colors = [0.0] * n
  k = 0
  for i in range(0,len(data)):
    for j in data[i]:
      xdata[k] = j[0]
      ydata[k] = j[1]
      colors[k] = cm.hsv(i/cnum)
      k = k + 1
  plt.scatter(xdata, ydata, c = colors, **kwargs)
  plt.show()


